from dataclasses import dataclass, field

from sm_blueprint_lib.bases.controllers.basecontroller import BaseController


@dataclass
class BaseLogicController(BaseController):
    """Base class for Logic parts' Controllers (mostly Logic Gate and Timer)
    """
    active: bool = field(kw_only=True, default=False)



from typing import Sequence
from numpy import ndarray
from sm_blueprint_lib import get_bits_required, check_pos, connect, num_to_bit_list
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.pos import Pos


def counter(bp: Blueprint,
            bit_length: int,
            pos: Pos | Sequence = (0, 0, 0),
            precreated_swxors=None, precreated_ands=None, precreated_count=None):
    pos = check_pos(pos)
    arr = ndarray((bit_length, 2), dtype=LogicGate)
    count = precreated_count if precreated_count is not None else LogicGate(
        pos + (-1, 1, 0), "FF0000", 1)
    for x in range(bit_length):
        arr[x, :] = [
            precreated_swxors[x] if precreated_swxors is not None else LogicGate(
                pos + (x, 0, 0), "0000FF", 2),
            precreated_ands[x] if precreated_ands is not None else LogicGate(
                pos + (x, 1, 0), "000000"),
        ]
    if precreated_swxors is None:
        connect(arr[:, 0], arr[:, 0])
    connect(arr[:, 1], arr[:, 0])
    for x in range(bit_length):
        connect(arr[x, 0], arr[x+1:, 1])
    connect(count, arr[:, 1])

    if precreated_swxors is None:
        bp.add(arr[:, 0])
    if precreated_ands is None:
        bp.add(arr[:, 1])
    if precreated_count is None:
        bp.add(count)
    return arr, count


def counter_decrement(bp: Blueprint,
                      bit_length: int,
                      pos: Pos | Sequence = (0, 0, 0),
                      precreated_swxors=None, precreated_nors=None, precreated_count_nor=None):
    pos = check_pos(pos)
    arr = ndarray((bit_length, 2), dtype=LogicGate)
    count = precreated_count_nor if precreated_count_nor is not None else LogicGate(
        pos + (-1, 1, 0), "FF0000", 4)
    for x in range(bit_length):
        arr[x, :] = [
            precreated_swxors[x] if precreated_swxors is not None else LogicGate(
                pos + (x, 0, 0), "0000FF", 2),
            precreated_nors[x] if precreated_nors is not None else LogicGate(
                pos + (x, 1, 0), "000000", 4),
        ]
    if precreated_swxors is None:
        connect(arr[:, 0], arr[:, 0])
    connect(arr[:, 1], arr[:, 0])
    for x in range(bit_length):
        connect(arr[x, 0], arr[x+1:, 1])
    connect(count, arr[:, 1])

    if precreated_swxors is None:
        bp.add(arr[:, 0])
    if precreated_nors is None:
        bp.add(arr[:, 1])
    if precreated_count_nor is None:
        bp.add(count)
    return arr, count



from dataclasses import dataclass, field
from typing import Any, Optional

from sm_blueprint_lib.constants import get_new_id
from sm_blueprint_lib.id import ID


@dataclass
class BaseController:
    """Base class for controller objects (used in interactable parts and so)
    """
    controllers: Optional[list[ID]] = field(kw_only=True, default=None)
    id: int = field(kw_only=True, default_factory=get_new_id)
    joints: Optional[list[ID]] = field(kw_only=True, default=None)

    def __post_init__(self):
        try:
            self.controllers = [ID(**c) for c in self.controllers]
        except TypeError:
            pass


from typing import Sequence
from numpy import ndarray
from sm_blueprint_lib import check_pos
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.pos import Pos
from sm_blueprint_lib.prebuilds.counter import counter, counter_decrement


def register(bp: Blueprint,
             bit_length: int,
             OE=True,
             pos: Pos | Sequence = (0, 0, 0)):
    pos = check_pos(pos)
    write = LogicGate(pos + (-1, 2 if OE else 1, 0), "FF0000", 1)
    if OE:
        output_enable = LogicGate(pos + (-1, 0, 0), "FF0000", 1)
        arr = ndarray((bit_length, 4), LogicGate)
        for x in range(bit_length):
            arr[x] = [
                l0 := LogicGate(pos + (x, 0, 0), "000000"),
                l1 := LogicGate(pos + (x, 1, 0), "0000FF", 2),
                l2 := LogicGate(pos + (x, 2, 0), "000000"),
                l3 := LogicGate(pos + (x, 3, 0), "FF0000", 2),
            ]
            l3.connect(l2).connect(l1).connect(l1).connect(l3)
            l1.connect(l0)
            write.connect(l2)
            output_enable.connect(l0)
        bp.add(output_enable)
    else:
        arr = ndarray((bit_length, 3), LogicGate)
        for x in range(bit_length):
            arr[x] = [
                l0 := LogicGate(pos + (x, 0, 0), "0000FF", 2),
                l1 := LogicGate(pos + (x, 1, 0), "000000"),
                l2 := LogicGate(pos + (x, 2, 0), "FF0000", 2),
            ]
            l2.connect(l1).connect(l0).connect(l0).connect(l2)
            write.connect(l1)
    bp.add(arr, write)
    if OE:
        return arr, write, output_enable
    else:
        return arr, write


def counter_register(bp: Blueprint,
                     bit_length: int,
                     OE=True,
                     with_increment=True,
                     with_decrement=True,
                     pos: Pos | Sequence = (0, 0, 0)):
    pos = check_pos(pos)
    r = register(bp, bit_length, OE, pos)

    if with_decrement:
        cdec = counter_decrement(bp,
                                 bit_length=bit_length,
                                 pos=pos+(0, OE, 1),
                                 precreated_swxors=r[0][:, int(OE)])
    if with_increment:
        cinc = counter(bp,
                       bit_length=bit_length,
                       pos=pos+(0, OE, 1+with_decrement),
                       precreated_swxors=r[0][:, int(OE)])
    if with_increment and with_decrement:
        return r, cinc[0][:, 1], cinc[1], cdec[0][:, 1], cdec[1]
    elif with_increment:
        return r, cinc[0][:, 1], cinc[1]
    else:
        return r, cdec[0][:, 1], cdec[1]



from typing import Sequence
from numpy import ndarray
from sm_blueprint_lib import check_pos, connect
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.pos import Pos


def comparator(bp: Blueprint,
               bit_length: int,
               pos: Pos | Sequence = (0, 0, 0)):
    pos = check_pos(pos)
    inputs = ndarray((bit_length, 4), dtype=LogicGate)
    ands_2 = ndarray((bit_length, 2), dtype=LogicGate)
    nors_3 = ndarray(bit_length, dtype=LogicGate)
    ands_4 = ndarray(2 * (bit_length - 1) + 1, dtype=LogicGate)
    ors_5 = ndarray(2, dtype=LogicGate)
    inputs[:, 0] = [
        LogicGate(pos+(x, -1, 0), "FF0000", 4)
        for x in range(bit_length)
    ]
    inputs[:, 1] = [
        LogicGate(pos+(x, 0, 0), "FF0000", 1)
        for x in range(bit_length)
    ]
    inputs[:, 2] = [
        LogicGate(pos+(x, -1, 2), "FF0000", 4)
        for x in range(bit_length)
    ]
    inputs[:, 3] = [
        LogicGate(pos+(x, 0, 2), "FF0000", 1)
        for x in range(bit_length)
    ]
    for x in range(bit_length):
        ands_2[x, :] = [
            LogicGate(pos+(x, -2, 0), "000000"),
            LogicGate(pos+(x, -2, 1), "000000"),
        ]
        nors_3[x] = LogicGate(pos+(x, -2, 2), "000000", 4)

    for n in range(2 * (bit_length - 1) + 1):
        if n == 0:
            ands_4[n] = LogicGate(pos+(1, -4, 0), "0000FF")
        else:
            ands_4[n] = LogicGate(pos+((n-1)//2, -3, (n-1) % 2), "000000")

    ors_5[:] = [
        LogicGate(pos+(0, -4, 0), "0000FF", 1),
        LogicGate(pos+(2, -4, 0), "0000FF", 1),
    ]

    connect(inputs[:, 0], ands_2[:, 1])
    connect(inputs[:, 1], ands_2[:, 0])
    connect(inputs[:, 2], ands_2[:, 0])
    connect(inputs[:, 3], ands_2[:, 1])

    connect(ands_2, nors_3)

    connect(ands_2.flat, ands_4[1:])
    for x in range(bit_length):
        connect(nors_3[x], ands_4[:2*x+1])

    connect(ands_4[1::2], ors_5[0])
    connect(ands_4[2::2], ors_5[1])
    connect(ands_2[-1, 0], ors_5[0])
    connect(ands_2[-1, 1], ors_5[1])

    bp.add(inputs, ands_2, nors_3, ands_4, ors_5)

    return inputs, ands_2, nors_3, ands_4, ors_5



from dataclasses import dataclass, field

from sm_blueprint_lib.bases.controllers.baselogiccontroller import BaseLogicController


@dataclass
class TimerController(BaseLogicController):
    """Timer's Controller
    """
    seconds: int
    ticks: int



from dataclasses import dataclass
from typing import Sequence


@dataclass
class Pos:
    """Class that represents the position of a block (x, y, z)
    """
    x: int
    y: int
    z: int

    def __add__(self, o: "Pos" | Sequence):
        if isinstance(o, Pos):
            return Pos(self.x + o.x, self.y + o.y, self.z + o.z)
        return Pos(self.x + o[0], self.y + o[1], self.z + o[2])



from typing import Sequence
from numpy import ndarray
from sm_blueprint_lib import get_bits_required, check_pos, connect, num_to_bit_list
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.pos import Pos


def clock40hz(bp: Blueprint, bit_length: int, pos: Pos | Sequence = (0, 0, 0)):
    pos = check_pos(pos)

    arr = ndarray((bit_length, 2), dtype=LogicGate)
    for x in range(bit_length):
        arr[x, :] = [
            LogicGate(pos + (x, 0, 0), "0000FF", 2 if x else 4),
            LogicGate(pos + (x, 1, 0), "000000", 0),
        ]

    connect(arr[:, 0], arr[:, 0])
    connect(arr[1:, 1], arr[1:, 0])
    connect(arr[0, 0], arr[0, 1])
    connect(arr[0, 1], arr[1:, 1])
    for x in range(1, bit_length):
        connect(arr[x, 0], arr[x+1:, 1])

    bp.add(arr)

    return arr



from typing import Sequence
from numpy import ndarray
from sm_blueprint_lib import get_bits_required, check_pos, connect, num_to_bit_list
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.pos import Pos


def decoder(bp: Blueprint, num_address: int, pos: Pos | Sequence = (0, 0, 0),
            precreated_inputs_binary=None, precreated_outputs=None, precreated_output_enable=None,
            address_divisor=1, with_enable=True):
    pos = check_pos(pos)

    if precreated_inputs_binary is not None:
        inputs_binary = precreated_inputs_binary
    else:
        inputs_binary = ndarray(
            (get_bits_required(num_address), 2), dtype=LogicGate)
    if precreated_outputs is not None:
        outputs = precreated_outputs
    else:
        outputs = [LogicGate(pos + (x+1, 0, 0), "0000FF")
                   for x in range(num_address)]
    if with_enable:
        if precreated_output_enable is not None:
            output_enable = precreated_output_enable
        else:
            output_enable = LogicGate(pos + (0, 0, 0), "FF0000", 1)

    if precreated_inputs_binary is None:
        for b in range(get_bits_required(num_address)):
            inputs_binary[b, :] = [
                LogicGate(pos + (b+1, 1, 0), "FF0000", 4),
                LogicGate(pos + (b+1, 2, 0), "FF0000", 1),
            ]

    for x in range(num_address):
        bit_mask = num_to_bit_list(
            x//address_divisor, get_bits_required(num_address//address_divisor))
        connect(inputs_binary[~bit_mask, 0], outputs[x])
        connect(inputs_binary[bit_mask, 1], outputs[x])
    if with_enable:
        connect(output_enable, outputs)

    if precreated_inputs_binary is None:
        bp.add(inputs_binary)
    if precreated_outputs is None:
        bp.add(outputs)
    if with_enable:
        if precreated_output_enable is None:
            bp.add(output_enable)

    if with_enable:
        return inputs_binary, outputs, output_enable
    else:
        return inputs_binary, outputs



from dataclasses import dataclass, field

from sm_blueprint_lib.bases.controllers.baselogiccontroller import BaseLogicController


@dataclass
class LogicGateController(BaseLogicController):
    """Logic Gate's Controller
    """
    mode: int = 0



from dataclasses import dataclass, field

from sm_blueprint_lib.bases.parts.basepart import BasePart
from sm_blueprint_lib.constants import SHAPEID


@dataclass
class Body:
    childs: list[BasePart] = field(default_factory=list)

    def __post_init__(self):
        self.childs = [SHAPEID.SHAPEID_TO_CLASS[child["shapeId"]](**child)
                       for child in self.childs]



class SHAPEID:
    BARRIER_BLOCK = "09ca2713-28ee-4119-9622-e85490034758"
    LOGIC_GATE = "9f0f56e8-2c31-4d83-996c-d00a9b296c3f"
    TIMER = "8f7fd0e7-c46e-4944-a414-7ce2437bb30f"
    SENSOR5 = "20dcd41c-0a11-4668-9b00-97f278ce21af"
    SWITCH = "7cf717d7-d167-4f2d-a6e7-6b2c70aa3986"
    SHAPEID_TO_CLASS = {}


class AXIS:
    DEFAULT_XAXIS = 1
    DEFAULT_ZAXIS = 3


class VERSION:
    BLUEPRINT_VERSION = 4


TICKS_PER_SECOND = 40

__global_id_counter = 0
"""Atempting to modify this global variable may cause to break your blueprints lol.
"""


def get_new_id():
    global __global_id_counter
    __global_id_counter = (new_id := __global_id_counter) + 1
    return new_id



from typing import Sequence
from numpy import ndarray
from sm_blueprint_lib import get_bits_required, check_pos, connect, num_to_bit_list
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.pos import Pos


def barrel_shifter(bp: Blueprint, bit_length: int, num_bit_shift: int, pos: Pos | Sequence = (0, 0, 0)):
    pos = check_pos(pos)
    inputs = [LogicGate(pos + (x, 0, 0), "FF0000", 1)
              for x in range(bit_length)]
    inputs_shift_binary = ndarray((num_bit_shift, 2), dtype=LogicGate)
    # always_off = LogicGate(pos + (-1, -1, 0), "000000")

    for x in range(num_bit_shift):
        inputs_shift_binary[x, :] = [
            LogicGate(pos + (x-num_bit_shift, -1, 0), "FF0000", 4),
            LogicGate(pos + (x-num_bit_shift, 0, 0), "FF0000", 1),
        ]

    arr = [
        [
            [LogicGate(pos+(x, (-y-1), 0), "000000")
             for x in range(bit_length + (2**y - 1 if y != num_bit_shift - 1 else 0))]
            for y in range(num_bit_shift)
        ],
        [
            [LogicGate(pos+((x - (2**y if y == num_bit_shift-1 else 0)), (-y-1), 1), "000000")
             for x in range(2**y, bit_length + 2**(y+1) - 1 - (2**y if y == num_bit_shift-1 else 0))]
            for y in range(num_bit_shift)
        ],
        [
            [LogicGate(pos+(x, (-y-1), 2), "000000" if y != num_bit_shift - 1 else "0000FF", 1)
             for x in range(bit_length - 1 + (2**(y+1) if y != num_bit_shift-1 else 1))]
            for y in range(num_bit_shift)
        ],
    ]

    connect(inputs, arr[0][0])
    connect(inputs, arr[1][0])
    for y in range(num_bit_shift):
        connect(arr[0][y], arr[2][y])
    for y in range(num_bit_shift-1):
        connect(arr[2][y], arr[0][y+1])
        connect(arr[1][y], arr[2][y][2**y:])
    for y in range(num_bit_shift-2):
        connect(arr[2][y], arr[1][y+1])
    connect(arr[2][-2][2**(num_bit_shift-1):], arr[1][-1])
    connect(arr[1][-1], arr[2][-1])
    for y in range(num_bit_shift):
        connect(inputs_shift_binary[y, 0], arr[0][y])
        connect(inputs_shift_binary[y, 1], arr[1][y])

    bp.add(inputs, inputs_shift_binary, arr)
    # arr = ndarray((bit_length, num_bit_shift, 3), dtype=LogicGate)
    # for x in range(num_bit_shift):
    #     inputs_shift_binary[x, :] = [
    #         LogicGate(pos + (x+bit_length+1, -1, 0), "FF0000", 4),
    #         LogicGate(pos + (x+bit_length+1, 0, 0), "FF0000", 1),
    #     ]
    # for x in range(bit_length):
    #     for y in range(num_bit_shift):
    #         arr[x, y, :] = [
    #             LogicGate(pos + (x, -y-1, 0), "000000"),
    #             LogicGate(pos + (x, -y-1, 1), "000000"),
    #             LogicGate(pos + (x, -y-1, 2), "000000"
    #                       if y + 1 < num_bit_shift else "0000FF", 1),
    #         ]
    # always_off.connect(always_off)

    # connect(inputs, arr[:, 0, 0])
    # connect(inputs, arr[1:, 0, 1])
    # connect(arr[:, :, 0], arr[:, :, 2])
    # connect(arr[:, :, 1], arr[:, :, 2])
    # connect(arr[:, :, 2], arr[:, 1:, 0])
    # for y in range(num_bit_shift):
    #     if y + 1 < num_bit_shift:
    #         connect(arr[:, y, 2], arr[2**(y+1):, y+1, 1])
    #     connect(always_off, arr[:2**y, y, 1])
    #     connect(inputs_shift_binary[y, 0], arr[:, y, 0])
    #     connect(inputs_shift_binary[y, 1], arr[:, y, 1])

    # bp.add(inputs, inputs_shift_binary, arr, always_off)

    # return inputs, inputs_shift_binary, arr, always_off



from dataclasses import dataclass, field

from sm_blueprint_lib.bases.controllers.basecontroller import BaseController


@dataclass
class SensorController(BaseController):
    """Sensor's Controller
    """
    audioEnable: bool
    buttonMode: bool
    color: str
    colorMode: bool
    range: int

    def __post_init__(self):
        super().__post_init__()
        # if color given as (r, g, b) then convert to hex string
        if not isinstance(self.color, str):
            self.color = "%02X%02X%02X" % (
                self.color[0], self.color[1], self.color[2])
        self.audioEnable = bool(self.audioEnable)
        self.buttonMode = bool(self.buttonMode)
        self.colorMode = bool(self.colorMode)



import os
from random import randint
from math import sin
from typing import Sequence

from numpy import array
from sm_blueprint_lib import Blueprint, connect, num_to_bit_list, save_blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.pos import Pos
from sm_blueprint_lib.prebuilds.adder import simple_adder_subtractor
from sm_blueprint_lib.prebuilds.barrel_shifter import barrel_shifter
from sm_blueprint_lib.prebuilds.clock40hz import clock40hz
from sm_blueprint_lib.prebuilds.comparator import comparator
from sm_blueprint_lib.prebuilds.counter import counter
from sm_blueprint_lib.prebuilds.distance_sensor import distance_sensor, distance_sensor_raycast
from sm_blueprint_lib.prebuilds.ram import ram
from sm_blueprint_lib.prebuilds.decoder import decoder
from sm_blueprint_lib.prebuilds.register import counter_register, register
from sm_blueprint_lib.prebuilds.rom import rom
from sm_blueprint_lib.prebuilds.timer_ram_multiclient import timer_ram_multiclient
from sm_blueprint_lib.prebuilds.timer_ram_cached import timer_ram_cached


bp = Blueprint()

timer_ram_cached(bp,
                 bit_length=8,
                 num_address_per_cache=8,
                 num_caches=4,
                 num_timer_banks=4,
                 num_caches_per_bank=128)

# comparator(bp, bit_length=32)

# simple_adder_subtractor(bp, bit_length=32)

# counter_register(bp,
#                  bit_length=6,
#                  OE=True,
#                  with_increment=True,
#                  with_decrement=True)


# clock40hz(bp, 10)

# timer_ram_multiclient(bp, bit_length=32, num_address=16, num_clients=1)

# timer_ram_cached(bp,
#                  bit_length=32,
#                  num_address_per_cache=4,
#                  num_caches=4,
#                  num_timer_banks=4,
#                  num_caches_per_bank=1024)


# barrel_shifter(bp,
#                bit_length=32,
#                num_bit_shift=6)

# distance_sensor(bp,
#                 range(1, 9, 1))

# distance_sensor_raycast(bp,
#                         range(1, 21, 2))


# # Initial instructions rom
# micro_ins = [100,101,102,103,104,105,106,107,108]
# rom1 = rom(bp, pos=(-30, -30, 0),
#            page_size=(32, 8),
#            data=micro_ins)


# register(bp, pos=(15,-5,0), bit_length=24, OE=False)

# decoder(bp, pos=(30, 5, 0), num_address=16)

# ram(bp, pos=(20,5,0),
#     bit_length=8,
#     num_address=8)

# reg = register(bp, pos=(0, -10, 0), bit_length=32)
# c = counter(bp, pos=(0, -9, 1), bit_length=32,
#             precreated_swxors=reg[0][:, 1])


print(f"Prebuild size: {len(bp.bodies[0].childs)} parts")
path = r"C:\Users\mauri\AppData\Roaming\Axolot Games\Scrap Mechanic\User\User_76561198400983548\Blueprints\c35f6e4e-52cb-4b00-8afa-f0ffd3fbb012\blueprint.json"
save_blueprint(path, bp)



from typing import Sequence
from numpy import ndarray
from sm_blueprint_lib import get_bits_required, check_pos, connect, num_to_bit_list
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.pos import Pos


def simple_adder_subtractor(bp: Blueprint,
                            bit_length: int,
                            pos: Pos | Sequence = (0, 0, 0)):
    pos = check_pos(pos)
    input_a = [LogicGate(pos+(x, 0, 2), "FF0000", 1)
               for x in range(bit_length)]
    input_b = [LogicGate(pos+(x, 0, 0), "FF0000", 2)
               for x in range(bit_length)]
    and_0 = [LogicGate(pos+(x, -1, 2), "000000", 0)
             for x in range(bit_length)]
    xor_0 = [LogicGate(pos+(x, -1, 0), "000000", 2)
             for x in range(bit_length)]
    or_0 = [LogicGate(pos+((x, -2, 0) if x + 1 != bit_length else (x+1, -2, 2)), "000000" if x + 1 != bit_length else "0000FF", 1)
            for x in range(bit_length)]
    and_1 = [LogicGate(pos+(x, -2, 1), "000000", 0)
             for x in range(bit_length)]
    xor_1 = [LogicGate(pos+(x, -2, 2), "0000FF", 2)
             for x in range(bit_length)]
    carry_in = LogicGate(pos+(-1, -2, 0), "FF0000", 1)

    connect(input_a, and_0)
    connect(input_a, xor_0)
    connect(input_b, and_0)
    connect(input_b, xor_0)
    connect(and_0, or_0)
    connect(xor_0, and_1)
    connect(xor_0, xor_1)
    connect(and_1, or_0)
    connect(or_0, and_1[1:])
    connect(or_0, xor_1[1:])
    connect(carry_in, and_1[0])
    connect(carry_in, xor_1[0])
    connect(carry_in, input_b)

    bp.add(input_a, input_b, and_0, xor_0, or_0, and_1, xor_1, carry_in)
    return input_a, input_b, and_0, xor_0, or_0, and_1, xor_1, carry_in



from dataclasses import dataclass, field

from sm_blueprint_lib.bases.parts.basepart import BasePart
from sm_blueprint_lib.body import Body
from sm_blueprint_lib.constants import VERSION


@dataclass
class Blueprint:
    bodies: list[Body] = field(default_factory=lambda: [Body()])
    version: int = VERSION.BLUEPRINT_VERSION

    def __post_init__(self):
        try:
            self.bodies = [Body(**body) for body in self.bodies]
        except TypeError:
            pass

    def add(self, *obj, body=0):
        """Adds the object(s) to the blueprint.

        Args:
            obj (Any): Can be a instance of BasePart or a subclass. It also can be any nested iterable of instances (list of parts, list of lists of parts, etc).
            body (int, optional): Specify in which blueprint's body the object will be placed. Defaults to 0.
        """
        for subobj in obj:
            if isinstance(subobj, BasePart):
                self.bodies[body].childs.append(subobj)
            else:
                for subsubobj in subobj:
                    self.add(subsubobj, body=body)



from typing import Sequence
from numpy import array, ndarray
from sm_blueprint_lib import get_bits_required, check_pos, connect, num_to_bit_list
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.parts.sensor import Sensor5
from sm_blueprint_lib.parts.timer import Timer
from sm_blueprint_lib.pos import Pos


def distance_sensor(bp: Blueprint,
                    sensor_range: range,
                    pos: Pos | Sequence = (0, 0, 0)):
    pos = check_pos(pos)
    bit_length = get_bits_required(len(sensor_range))
    out = [LogicGate(pos + (x, 0, 0), "0000FF", 2) for x in range(bit_length)]
    out = array(out, dtype=LogicGate)
    sensors = [Sensor5(pos + (bit_length, -1, 1), "000000",
                       (False, True, "FFFFFF", False, x), xaxis=-3, zaxis=1) for x in sensor_range]

    current_output_state = 0
    for target in reversed(range(len(sensor_range))):
        difference = current_output_state ^ (target+1)
        if difference:
            bit_mask = num_to_bit_list(difference, bit_length)
            connect(sensors[target], out[bit_mask])
            current_output_state ^= difference

    bp.add(out, sensors)
    return out, sensors


def distance_sensor_raycast(bp: Blueprint,
                            sensor_range: range,
                            pos: Pos | Sequence = (0, 0, 0)):
    pos = check_pos(pos)
    bit_length = len(sensor_range)
    out = [LogicGate(pos + (x, 0, 0), "0000FF", 2) for x in range(bit_length)]
    sensors = [Sensor5(pos + (bit_length, -1, 0), "000000",
                       (False, True, "FFFFFF", False, x), xaxis=1, zaxis=3) for x in sensor_range]

    connect(sensors, out)

    bp.add(out, sensors)
    return out, sensors



from dataclasses import asdict
from json import load, dump, loads, dumps
from math import ceil, log2

from numpy import ndarray

from sm_blueprint_lib.bases.parts.baseinteractablepart import BaseInteractablePart
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.barrierblock import BarrierBlock
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.parts.timer import Timer
from sm_blueprint_lib.pos import Pos


def load_blueprint(path):
    """Load a blueprint from a path file (normally a blueprint.json).

    Args:
        path (str): The path to the json file.

    Returns:
        Blueprint: The loaded blueprint.
    """
    with open(path) as fp:
        return Blueprint(**load(fp))


def save_blueprint(path, bp: Blueprint):
    """Save a blueprint to a file (normally a blueprint.json).

    Args:
        path (str): The path to save the json file.
        bp (Blueprint): The blueprint to be saved.
    """
    with open(path, mode="w") as fp:
        return dump(asdict(bp), fp, sort_keys=True, separators=(',', ':'))


def load_string(str):
    """Load a blueprint from a json string.

    Args:
        str (str): The string to be loaded.

    Returns:
        Blueprint: The loaded blueprint.
    """
    return Blueprint(**loads(str))


def dump_string(bp: Blueprint):
    """Dump a blueprint into a json-formatted string.

    Args:
        bp (Blueprint): The blueprint to be dumped.

    Returns:
        str: The json-formatted string.
    """
    return dumps(asdict(bp), sort_keys=True, separators=(',', ':'))


def connect(_from, _to, *, parallel=True):
    """Connect interactable parts together, recursively.

    Args:
        _from (Any): Must be an instance of BaseInteractablePart or a subclass.
        Also it can be any nested iterable of instances (list of parts, list of lists of parts, etc).
        _to (Any): Must be an instance of BaseInteractablePart or a subclass.
        Also it can be any nested iterable of instances (list of parts, list of lists of parts, etc).
        parallel (bool, optional): Defines the behaviour of the connections in the following way:

        With parallel=False, everything connects to everything:
            from1 🔀 to1

            from2 🔀 to2

        With parallel=True, every row is connected respectively:
            from1 → to1

            from2 → to2

        Also, if the dimensions does not match it tries to adapt (many to one, one to many, etc)

        Defaults to True.
    """
    if isinstance(_from, BaseInteractablePart) and isinstance(_to, BaseInteractablePart):
        _from.connect(_to)
        return
    if parallel:    # Try connect things row-by-row if possible (one to one, one to many, many to many)
        if not isinstance(_from, BaseInteractablePart) and not isinstance(_to, BaseInteractablePart):   # Assume both are sequence of parts
            for subfrom, subto in zip(_from, _to):
                connect(subfrom, subto, parallel=parallel)
        elif not isinstance(_from, BaseInteractablePart):   # Assume _from is a sequence of parts
            for subfrom in _from:
                connect(subfrom, _to, parallel=parallel)
        else:                                               # Assume _to is a sequence of parts
            for subto in _to:
                connect(_from, subto, parallel=parallel)
    else:           # Just connect everything to everything lol
        if not isinstance(_from, BaseInteractablePart) and not isinstance(_to, BaseInteractablePart):   # Assume both are sequence of parts
            for subfrom in _from:
                for subto in _to:
                    connect(subfrom, subto, parallel=parallel)
        elif not isinstance(_from, BaseInteractablePart):   # Assume _from is a sequence of parts
            for subfrom in _from:
                connect(subfrom, _to, parallel=parallel)
        else:                                               # Assume _to is a sequence of parts
            for subto in _to:
                connect(_from, subto, parallel=parallel)


def check_pos(pos):
    if not isinstance(pos, Pos):
        pos = Pos(*pos)
    return pos


def get_bits_required(number: int | float):
    """Calculates how many bits are required to store this number.

    Args:
        number (int | float): The target number.
    """
    return ceil(log2(number))


def num_to_bit_list(number: int, bit_length: int):
    """Converts a number to a numpy array of its bits.

    Args:
        number (int): The number to convert.
        bit_length (int): The number of bits the list will have.
    """
    output = ndarray(bit_length, dtype=bool)
    for b in range(bit_length):
        output[b] = bool((number >> b) & 1)
    return output



from typing import Sequence
from numpy import ndarray
from sm_blueprint_lib import get_bits_required, check_pos, connect
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.parts.timer import Timer
from sm_blueprint_lib.pos import Pos
from sm_blueprint_lib.prebuilds.clock40hz import clock40hz
from sm_blueprint_lib.prebuilds.decoder import decoder
from sm_blueprint_lib.prebuilds.ram import ram
from sm_blueprint_lib.prebuilds.register import register


def timer_ram_multiclient(bp: Blueprint, bit_length: int, num_address: int, num_clients=1, pos: Pos | Sequence = (0, 0, 0)):
    initial_part_count = len(bp.bodies[0].childs)
    pos = check_pos(pos)

    clock = clock40hz(
        bp, get_bits_required(num_address), pos + (bit_length, 0, 0))

    timers = ndarray((bit_length, 3), dtype=object)
    timer_read_write_enable = LogicGate(pos + (-1, 1, 0), "FF0000", 4)
    for x in range(bit_length):
        # timers[x] = [
        #     t0 := Timer(pos + (x % 8, 0, 2*(x//8)), "000000", ((num_address-3) // 40, (num_address-3) % 40)),
        #     l0 := LogicGate(pos + (x % 8, 1, 2*(x//8)), "000000"),
        #     l1 := LogicGate(pos + (x % 8, 1, 2*(x//8)+1), "000000", 1),
        # ]
        timers[x] = [
            t0 := Timer(pos + (x, 0, 0), "000000", ((num_address-3) // 40, (num_address-3) % 40)),
            l0 := LogicGate(pos + (x, 1, 0), "000000"),
            l1 := LogicGate(pos + (x, 1, 1), "000000", 1),
        ]
        t0.connect(l0).connect(l1).connect(t0)
    connect(timer_read_write_enable, timers[:, 1])

    for z in range(num_clients):
        cpos = pos + (0, 0, z*2)
        data = register(bp, bit_length, pos=cpos + (0, 4, 0))
        timers_readers = [LogicGate(cpos+(x, 3, 0), "000000")
                          for x in range(bit_length)]
        timers_readers_enable = LogicGate(cpos+(-1, 3, 0), "FF0000", 1)
        timers_readers_buffer = LogicGate(cpos+(-1, 5, 0), "000000")
        address = register(bp, get_bits_required(
            num_address), pos=cpos + (bit_length+1, 5, 0), OE=False)
        post_rw = register(bp, 2,
                           pos=cpos + (bit_length+1+get_bits_required(num_address)+1, 5, 0), OE=False)
        rw_inv = [
            LogicGate(
                cpos + (bit_length+1+get_bits_required(num_address)+1, 4, 0), "000000", 3),
            LogicGate(
                cpos + (bit_length+1+get_bits_required(num_address)+2, 4, 0), "000000")
        ]
        comparator = [LogicGate(cpos+(x+bit_length+1, 4, 0), "000000", 5)
                      for x in range(get_bits_required(num_address))]
        comparator_write = LogicGate(cpos+(bit_length+1, 3, 0), "000000")
        comparator_read = LogicGate(cpos+(bit_length+2, 3, 0), "000000")
        connect(post_rw[0][0, 0], rw_inv)
        connect(address[0][:, 0], comparator)
        connect(clock[:, 0], comparator)
        connect(comparator, comparator_write)
        connect(comparator, comparator_read)
        connect(rw_inv[0], comparator_write)
        connect(rw_inv[1], comparator_read)
        connect(post_rw[0][1, 0], comparator_write)
        connect(post_rw[0][1, 0], comparator_read)
        connect(comparator_write, timer_read_write_enable)
        connect(comparator_write, data[2])
        connect(data[0][:, 0], timers[:, 2])
        connect(comparator_write, post_rw[1])
        connect(comparator_read, post_rw[1])
        connect(comparator_read, timers_readers_enable)
        connect(timers_readers_enable, timers_readers_buffer)
        connect(timers_readers_buffer, data[1])
        connect(timers_readers_enable, timers_readers)
        connect(timers[:, 0], timers_readers)
        connect(timers_readers, data[0][:, 3])
        bp.add(comparator, comparator_write, comparator_read, rw_inv,
               timers_readers, timers_readers_enable, timers_readers_buffer)

    bp.add(timers, timer_read_write_enable)

    print(f"Timer ram: {bit_length*num_address} bits ({bit_length*num_address/8} bytes), "
          f"Max time delay: {num_address} ticks ({num_address/40} seconds), "
          f"Bits per part: {bit_length*num_address / (len(bp.bodies[0].childs)-initial_part_count):.2f} bits/part")



from itertools import cycle
from typing import Sequence
from numpy import ndarray
from sm_blueprint_lib import get_bits_required, check_pos, connect
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.pos import Pos
from sm_blueprint_lib.prebuilds.decoder import decoder


def ram(bp: Blueprint, bit_length: int, num_address: int, pos: Pos | Sequence = (0, 0, 0),
        address_divisor=1,
        pre_arr=None, pre_inputs=None, pre_outputs=None):
    pos = check_pos(pos)

    arr = pre_arr if pre_arr is not None else ndarray(
        (bit_length, num_address, 4), dtype=LogicGate)
    writers = [LogicGate(pos + (-1, 2, y), "000000", xaxis=1, zaxis=3)
               for y in range(num_address)]
    readers = [LogicGate(pos + (-1, 0, y), "000000", xaxis=1, zaxis=3)
               for y in range(num_address)]
    inputs = pre_inputs if pre_inputs is not None else [LogicGate(pos + (x, 1, -1), "FF0000", 1, xaxis=1, zaxis=3)
                                                        for x in range(bit_length)]
    outputs = pre_outputs if pre_outputs is not None else [LogicGate(pos + (x, 1, -2), "0000FF", 1, xaxis=1, zaxis=3)
                                                           for x in range(bit_length)]
    writers_binary = ndarray(
        (get_bits_required(num_address//address_divisor), 2), dtype=LogicGate)
    write_enable = LogicGate(pos + (-1, 2, -1), "FF0000", 1, xaxis=1, zaxis=3)
    readers_binary = ndarray(
        (get_bits_required(num_address//address_divisor), 2), dtype=LogicGate)
    read_enable = LogicGate(
        pos + (-1, 0, -1), "FF0000", 1, xaxis=1, zaxis=3)

    if pre_arr is None:
        for x in range(bit_length):
            for y in range(num_address):
                # if pre_arr is None:
                arr[x, y, :] = [
                    # arr.flat[y*num_address + x] = [
                    l3 := LogicGate(pos + (x, 1, y+1), "000000", xaxis=1, zaxis=-3),
                    l2 := LogicGate(pos + (x, 2, y), "FF0000", 2),
                    l1 := LogicGate(pos + (x, 3, y), "000000"),
                    l0 := LogicGate(pos + (x, 3, y), "0000FF", 2, xaxis=1, zaxis=3),
                ]
                l2.connect(l1).connect(l0).connect(l0).connect(l2)
                l0.connect(l3)
                # else:
                #     print(x, y, bit_length, num_address, arr.shape)
                #     l3, l2, l1, l0 = arr[x, y, :]
                #     # l3, l2, l1, l0 = arr.reshape(())[x, y, :]
                #     l2.connect(l1).connect(l0).connect(l2)
                #     l0.connect(l3)
    else:
        for l3, l2, l1, l0 in arr.reshape((arr.shape[0]*arr.shape[1], 4)):
            l2.connect(l1).connect(l0).connect(l2)
            l0.connect(l3)
    connect(writers, arr[:, :, 2].T)
    connect(readers, arr[:, :, 0].T)
    # connect(inputs, arr[:, :, 1])
    for i, a in zip(cycle(inputs), arr[:, :, 1].T.flat):
        connect(i, a)
    # connect(arr[:, :, 0], outputs)
    for a, o, in zip(arr[:, :, 0].T.flat, cycle(outputs)):
        connect(a, o)
    for y in range(get_bits_required(num_address//address_divisor)):
        writers_binary[y, :] = [
            LogicGate(pos + (-2, 2, y), "FF0000", 4, xaxis=1, zaxis=3),
            LogicGate(pos + (-3, 2, y), "FF0000", 1, xaxis=1, zaxis=3),
        ]
        readers_binary[y, :] = [
            LogicGate(pos + (-2, 0, y),
                      "FF0000", 4, xaxis=1, zaxis=3),
            LogicGate(pos + (-3, 0, y),
                      "FF0000", 1, xaxis=1, zaxis=3),
        ]
    decoder(bp, num_address, (0, 0, 0), writers_binary, writers,
            write_enable, address_divisor=address_divisor)
    decoder(bp, num_address, (0, 0, 0), readers_binary, readers,
            read_enable, address_divisor=address_divisor)

    if pre_arr is None:
        bp.add(arr)
    if pre_inputs is None:
        bp.add(inputs)
    if pre_outputs is None:
        bp.add(outputs)
    bp.add(writers, readers, writers_binary,
           write_enable, readers_binary, read_enable)

    return arr, inputs, outputs, writers_binary, writers, write_enable, readers_binary, readers, read_enable



import json
from pprint import pp


from sm_blueprint_lib import load_blueprint, save_blueprint, load_string
# from sm.bases.controllers.basecontroller import BaseController
# from sm.bases.controllers.baselogiccontroller import BaseLogicController
# from sm.bases.parts.baselogicpart import BaseLogicPart
# import sm.bases.parts.basepart
# from sm.bases.controllers.logicgatecontroller import LogicGateController
# from sm.blueprint import Blueprint
# import sm.bases.parts.baseboundablepart
# import sm.parts.barrierblock
# from dataclasses import asdict

from sm_blueprint_lib.constants import SHAPEID
from sm_blueprint_lib.parts.sensor import Sensor5
# from sm.bases.parts.baseinteractablepart import BaseInteractablePart
# from sm.parts.logicgate import LogicGate

path = r"C:\Users\mauri\AppData\Roaming\Axolot Games\Scrap Mechanic\User\User_76561198400983548\Blueprints\1d5e07ec-c750-4a24-ade9-0e278867ba81\blueprint.json"

with open(path) as fp:
    bp = json.load(fp)
    all_parts = bp["bodies"][0]["childs"]
    all_fields = set()
    print(len(all_parts))

    for part in all_parts:
        all_fields |= set(part.keys())
    print(f"{all_fields=}")

    common_fields = all_fields.copy()
    for part in all_parts:
        common_fields &= set(part.keys())
    print(f"{common_fields=}")

    fields_boundable = all_fields.copy()
    for part in all_parts:
        if part.get("bounds"):
            fields_boundable &= set(part.keys())
    print(f"{fields_boundable=}")

    parts_with_controller = [p for p in all_parts if p.get("controller")]
    all_fields_controller = set([(SHAPEID.SHAPEID_TO_CLASS.get(
        p["shapeId"]), *tuple(p["controller"].keys())) for p in parts_with_controller])
    common_controller = list(set(x) for x in all_fields_controller.copy())
    print("common_controller=", end="")
    pp(common_controller, width=100)

    a = list(set(x) for x in all_fields_controller.copy())
    all_controller = set()
    for s in a:
        all_controller |= s
    pp(all_controller, width=100)

    common_controller = all_controller.copy()
    for s in a:
        common_controller &= s
    pp(common_controller, width=100)


j = '{"bodies":[{"childs":[{"bounds":{"x":2,"y":2,"z":1},"color":"CE9E0C","pos":{"x":-13,"y":-7,"z":2},"shapeId":"09ca2713-28ee-4119-9622-e85490034758","xaxis":1,"zaxis":3},{"color":"DF7F01","controller":{"active":false,"controllers":[{"id":7227441}],"id":7227439,"joints":null,"seconds":2,"ticks":1},"pos":{"x":-12,"y":-5,"z":3},"shapeId":"8f7fd0e7-c46e-4944-a414-7ce2437bb30f","xaxis":1,"zaxis":-2},{"color":"DF7F01","controller":{"active":true,"controllers":[{"id":7227439}],"id":7227440,"joints":null,"mode":3},"pos":{"x":-12,"y":-6,"z":3},"shapeId":"9f0f56e8-2c31-4d83-996c-d00a9b296c3f","xaxis":1,"zaxis":-2},{"color":"DF7F01","controller":{"active":false,"controllers":[{"id":7227440}],"id":7227441,"joints":null,"mode":0},"pos":{"x":-13,"y":-6,"z":3},"shapeId":"9f0f56e8-2c31-4d83-996c-d00a9b296c3f","xaxis":1,"zaxis":-2}]}],"version":4}'
bp = load_string(j)
pp(bp)

pp(Sensor5((1,2,3), (255,255,255), (False, True, (255,0,0), False, 1)))

# print(BaseController())
# a = BaseInteractablePart("a", (1, 2, 3), (255, 0, 0), 1, 3)
# b = BaseInteractablePart("b", (2, 3, 4), (255, 0, 0), 1, 3)
# a.connect(b)
# print(a)
# print(b)
# print(BaseLogicController())
# print(BaseLogicPart("c", (3, 4, 5), (20, 30, 10), 1, 3))
# print(LogicGateController())
# l0 = LogicGate((0, 0, 0), (1, 2, 3), 1, 3)
# l1 = LogicGate((0, 0, 0), (1, 2, 3), 1, 3)
# l0.connect(l1)
# pp(l0)
# pp(l1)
# path = r"C:\Users\mauri\AppData\Roaming\Axolot Games\Scrap Mechanic\User\User_76561198400983548\Blueprints\68326f94-1bc7-446a-8e7c-63d4fb01d694\blueprint.json"
# bp = load_blueprint(path)
# pp(bp)


# p = sm.bases.parts.baseboundablepart.BaseBoundablePart("a",
#                                                  (0, 2, 3),
#                                                  (0, 0, 0),
#                                                  (1, 1, 1))
# pp(json.dumps(asdict(p)))

# j = '{"shapeId": "aa", "pos": {"x": 0, "y": 2, "z": 3}, "color": "000000", "xaxis": 0, "zaxis": 1}'
# p2 = sm.bases.parts.basepart.BasePart(**json.loads(j))
# pp(p2)

# j = '{"shapeId": "a", "pos": {"x": 0, "y": 2, "z": 3}, "color": "000000", "bounds": {"x": 1, "y": 1, "z": 1} }'
# p2 = sm.bases.parts.baseboundablepart.BaseBoundablePart(**json.loads(j))
# pp(p2)


# j = '''{
#                     "bounds": { "x": 1, "y": 1, "z": 1 },
#                     "color": "CE9E0C",
#                     "pos": { "x": 4, "y": 23, "z": 12 },
#                     "shapeId": "09ca2713-28ee-4119-9622-e85490034758",
#                     "xaxis": 1,
#                     "zaxis": 3
# }'''
# p2 = sm.parts.barrierblock.BarrierBlock(**json.loads(j))
# pp(p2)

# path = r"C:\Users\mauri\AppData\Roaming\Axolot Games\Scrap Mechanic\User\User_76561198400983548\Blueprints\68326f94-1bc7-446a-8e7c-63d4fb01d694\blueprint.json"
# bp = load_blueprint(path)
# pp(bp)
# pp(json.dumps(asdict(bp)))
# save_blueprint(path, bp)

# pp(sm.constants.SHAPEID.SHAPEID_TO_CLASS)

# pp(Blueprint())



from itertools import cycle
from typing import Sequence
from numpy import array, clip, ndarray
from sm_blueprint_lib import get_bits_required, check_pos, connect, num_to_bit_list
from sm_blueprint_lib.bases.parts.baseinteractablepart import BaseInteractablePart
from sm_blueprint_lib.bases.parts.baselogicpart import BaseLogicPart
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.constants import TICKS_PER_SECOND
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.parts.timer import Timer
from sm_blueprint_lib.pos import Pos
from sm_blueprint_lib.prebuilds.clock40hz import clock40hz
from sm_blueprint_lib.prebuilds.counter import counter
from sm_blueprint_lib.prebuilds.decoder import decoder
from sm_blueprint_lib.prebuilds.ram import ram
from sm_blueprint_lib.prebuilds.register import register


def timer_ram_cached(
        bp: Blueprint,
        bit_length: int,
        num_address_per_cache: int,
        num_caches: int,
        num_timer_banks: int,
        num_caches_per_bank: int,
        pos: Pos | Sequence = (0, 0, 0)):
    initial_part_count = len(bp.bodies[0].childs)
    total_bits = bit_length*num_address_per_cache*num_timer_banks*num_caches_per_bank
    cached_bits = bit_length*num_address_per_cache*num_caches
    cached_addresses = num_address_per_cache*num_caches
    banked_addresses = num_address_per_cache*num_timer_banks
    total_pages = num_timer_banks*num_caches_per_bank

    pos = check_pos(pos)

    clock = clock40hz(
        bp, get_bits_required(num_caches_per_bank), pos + (-get_bits_required(total_pages) - 5, 7, num_caches))

    cache = ram(bp,
                bit_length,
                cached_addresses,
                pos + (0, 5, 0))
    cache_arr = ndarray((bit_length, cached_addresses, 4), dtype=LogicGate)
    cache_full_page_in = ndarray(
        (bit_length, num_address_per_cache), dtype=LogicGate)
    cache_full_page_out = ndarray(
        (bit_length, num_address_per_cache), dtype=LogicGate)

    timer_bank = ndarray(
        (bit_length, banked_addresses, 5), dtype=BaseLogicPart)

    timer_bank_decoder_read = [LogicGate(
        pos + (bit_length*3, 8, z), "000000", xaxis=1, zaxis=3) for z in range(banked_addresses)]
    timer_bank_decoder_read_in = ndarray(
        (get_bits_required(num_timer_banks), 2), dtype=LogicGate)
    timer_bank_decoder_read_enable = LogicGate(
        pos + (bit_length*3, 8, -1), "FF0000", 1, xaxis=1, zaxis=3)

    timer_bank_decoder_write_keep = ndarray(
        (banked_addresses, 2), dtype=LogicGate)
    timer_bank_decoder_write_keep_in = ndarray(
        (get_bits_required(num_timer_banks), 4), dtype=LogicGate)
    timer_bank_decoder_write_keep_enable = (LogicGate(pos + (bit_length*3, 6, -1), "FF0000", 1, xaxis=1, zaxis=3),
                                            LogicGate(pos + (bit_length*3, 6, -2), "FF0000", 1, xaxis=1, zaxis=3))

    cache_pointer = counter(bp, get_bits_required(num_caches),
                            pos + (-get_bits_required(total_pages) - 5, 5, num_caches))

    # cache_table_offset_x = int(
    #     clip(get_bits_required(total_pages) - bit_length + 2, 0, None))
    cache_table = ram(bp,
                      bit_length=get_bits_required(total_pages) + 2,
                      num_address=num_caches,
                      #   pos=pos + (-cache_table_offset_x, 5, cached_addresses+2))
                      pos=pos + (-get_bits_required(total_pages) - 5, 5, 0))

    cache_table_lookup = ndarray(
        (get_bits_required(total_pages), num_caches), dtype=LogicGate)
    cache_table_lookup_in = [LogicGate(pos + (x-get_bits_required(total_pages) - 5, 4, -1),
                                       "ff0000", 1, xaxis=1, zaxis=3) for x in range(get_bits_required(total_pages))]
    cache_table_lookup_out = array([LogicGate(pos + (x-get_bits_required(total_pages) - 5, 4, -2),
                                   "0000ff", 1, xaxis=1, zaxis=3) for x in range(get_bits_required(num_caches)+1)], dtype=LogicGate)
    cache_table_lookup_dec = [LogicGate(pos + (-get_bits_required(
        total_pages) - 6, 4, z), "000000", xaxis=1, zaxis=3) for z in range(num_caches)]

    for x in range(bit_length):
        for y in range(cached_addresses):
            cache_arr[x, y, :] = [
                LogicGate(pos + (x, 3, y), "000000"),
                LogicGate(pos + (x, 4, y), "FF0000", 2),
                LogicGate(pos + (x, 5, y), "000000"),
                cache[0][x, y, 3]
            ]
        for y in range(num_address_per_cache):
            cache_full_page_in[x, y] = LogicGate(
                pos + (x, 2, y+1), "FF0000", 1, xaxis=1, zaxis=-3)
            cache_full_page_out[x, y] = LogicGate(
                pos + (x, 2, y+num_address_per_cache+1), "0000FF", 1, xaxis=1, zaxis=-3)
    for x in range(bit_length):
        for y in range(banked_addresses):
            timer_bank[x, y, :] = [
                LogicGate(pos + (x*2 + bit_length, 7, y), "FF0000"),
                LogicGate(pos + (x*2 + bit_length+1, 7, y), "0000FF"),
                LogicGate(pos + (x*2 + bit_length, 8, y), "000000"),
                LogicGate(pos + (x*2 + bit_length+1, 8, y), "000000", 1),
                Timer(pos + (x*2 + bit_length, 9, y+1), "0000FF",
                      divmod(num_caches_per_bank-3, TICKS_PER_SECOND), xaxis=-3, zaxis=-2)
            ]
    for x in range(get_bits_required(num_timer_banks)):
        timer_bank_decoder_read_in[x, :] = [
            LogicGate(pos + (bit_length*3+1, 8, x),
                      "FF0000", 4, xaxis=1, zaxis=3),
            LogicGate(pos + (bit_length*3+2, 8, x),
                      "FF0000", 1, xaxis=1, zaxis=3),
        ]
        timer_bank_decoder_write_keep_in[x, :] = [
            LogicGate(pos + (bit_length*3+2, 6, x),
                      "FF0000", 4, xaxis=1, zaxis=3),
            LogicGate(pos + (bit_length*3+3, 6, x),
                      "FF0000", 1, xaxis=1, zaxis=3),
            LogicGate(pos + (bit_length*3+4, 6, x),
                      "FF0000", 4, xaxis=1, zaxis=3),
            LogicGate(pos + (bit_length*3+5, 6, x),
                      "FF0000", 1, xaxis=1, zaxis=3),
        ]
    for z in range(banked_addresses):
        timer_bank_decoder_write_keep[z, :] = [
            LogicGate(pos + (bit_length*3, 6, z), "000000", xaxis=1, zaxis=3),
            LogicGate(pos + (bit_length*3+1, 6, z),
                      "000000", 3, xaxis=1, zaxis=3),
        ]
    for x in range(get_bits_required(total_pages)):
        for z in range(num_caches):
            cache_table_lookup[x, z] = LogicGate(
                pos + (x-get_bits_required(total_pages) - 5, 5, z+1), "000000", 5, xaxis=1, zaxis=-3)

    connect(timer_bank[:, :, 4], timer_bank[:, :, 2])
    connect(timer_bank[:, :, 2], timer_bank[:, :, 3])
    connect(timer_bank[:, :, 3], timer_bank[:, :, 4])
    connect(timer_bank[:, :, 4], timer_bank[:, :, 1])
    connect(timer_bank[:, :, 0], timer_bank[:, :, 3])

    decoder(bp, banked_addresses,
            precreated_outputs=timer_bank_decoder_read,
            precreated_inputs_binary=timer_bank_decoder_read_in,
            precreated_output_enable=timer_bank_decoder_read_enable,
            address_divisor=num_address_per_cache)

    decoder(bp, banked_addresses,
            precreated_outputs=timer_bank_decoder_write_keep[:, 0],
            precreated_inputs_binary=timer_bank_decoder_write_keep_in[:, :2],
            precreated_output_enable=timer_bank_decoder_write_keep_enable[0],
            address_divisor=num_address_per_cache)

    decoder(bp, banked_addresses,
            precreated_outputs=timer_bank_decoder_write_keep[:, 1],
            precreated_inputs_binary=timer_bank_decoder_write_keep_in[:, 2:],
            precreated_output_enable=timer_bank_decoder_write_keep_enable[1],
            address_divisor=num_address_per_cache)

    connect(timer_bank_decoder_read, timer_bank[:, :, 1].T)
    connect(timer_bank_decoder_write_keep[:, 0], timer_bank[:, :, 0].T)
    connect(timer_bank_decoder_write_keep[:, 1], timer_bank[:, :, 2].T)

    for t, c in zip(timer_bank[:, :, 1].T, cycle(cache_full_page_in.T)):
        connect(t, c)

    for c, t in zip(cycle(cache_full_page_out.T), timer_bank[:, :, 0].T):
        connect(c, t)

    connect(cache_table[0][:, :, -1], cache_table_lookup)
    connect(cache_table[0][-2, :, -1], cache_table_lookup_dec)
    connect(cache_table_lookup_in, cache_table_lookup)
    connect(cache_table_lookup.T, cache_table_lookup_dec)
    connect(cache_table_lookup_dec, cache_table_lookup_out[-1])
    for x in range(num_caches):
        bit_mask = num_to_bit_list(
            x, get_bits_required(num_caches)+1)
        connect(cache_table_lookup_dec[x], cache_table_lookup_out[bit_mask])

    ram(bp,
        bit_length=bit_length*num_address_per_cache,
        num_address=cached_addresses,
        pos=pos + (0, 1, 0),
        address_divisor=num_address_per_cache,
        pre_arr=cache_arr,
        pre_inputs=cache_full_page_in.T.flat,
        pre_outputs=cache_full_page_out.T.flat)

    control_pos = pos+(0, 15, 0)
    data_register = register(bp, bit_length=bit_length, pos=control_pos)
    address_register = register(
        bp, bit_length=get_bits_required(total_bits//bit_length), pos=control_pos+(bit_length+1, 0, 0))
    read_write_register = register(
        bp, bit_length=1, pos=control_pos+(bit_length+get_bits_required(total_bits//bit_length)+2, 0, 0))
    execute = LogicGate(pos=control_pos+(bit_length+get_bits_required(total_bits//bit_length)+3, 3, 0),
                        color="FF0000",
                        controller=1)
    execution_done = LogicGate(pos=control_pos+(bit_length+get_bits_required(total_bits//bit_length)+4, 3, 0),
                               color="0000FF",
                               controller=1)

    address_comparator = [
        # cache_table
        [LogicGate(pos + (-get_bits_required(total_pages) - 5 + x, 8, num_caches),
                   "000000",
                   5, xaxis=1, zaxis=3) for x in range(get_bits_required(num_caches_per_bank))],
        LogicGate(pos + (-get_bits_required(total_pages) - 5, 8, num_caches+1),
                  "000000", 0, xaxis=1, zaxis=3),
    ]
    connect(address_comparator[0], address_comparator[1])
    connect(clock[:, 0], address_comparator[0])
    connect(cache_table[2], address_comparator[0])

    row_start = control_pos+(-1, -1, 0)

    selectors = [
        [LogicGate(row_start+(x,5,0), "000000") for x in range(get_bits_required(total_pages))],
        [Timer(row_start+(x,6,0), "000000", (0,2)) for x in range(get_bits_required(total_pages))],
        [LogicGate(row_start+(x,7,0), "000000") for x in range(get_bits_required(total_pages))],
        [LogicGate(row_start+(x,8,0), "000000") for x in range(get_bits_required(num_caches))],
        # [Timer(row_start+(x,6,0), "000000", (0,0)) for x in range(get_bits_required(num_caches))],
        # [LogicGate(row_start+(x,7,0), "000000") for x in range(get_bits_required(num_caches))],
    ]
    connect(address_register[0][::-1, 1], selectors[0][::-1])
    connect(selectors[0], cache_table[1])
    connect(cache_table[2], selectors[1])
    connect(selectors[1], selectors[2])
    connect(selectors[2], cache_table[1])
    connect(cache_table_lookup_out, selectors[3])
    connect(selectors[3], cache_table[3])

    connect(address_register[0][::-1, 1], cache_table_lookup_in[::-1])
    connect(cache_table_lookup_out[-2::-1], cache[3][::-1])
    connect(cache_table_lookup_out[-2::-1], cache[6][::-1])
    connect(address_register[0][:get_bits_required(
        num_address_per_cache), 1], cache[3])
    connect(address_register[0][:get_bits_required(
        num_address_per_cache), 1], cache[6])
    connect(cache[2], data_register[0][:, -1])
    connect(data_register[0][:, 1], cache[1])
    connect(address_register[0][get_bits_required(
        num_address_per_cache):, 1], cache_table[1])
    # connect(cache_table_lookup_out, cache_table[3])
    connect(cache_pointer[0][:, 0], cache_table[6])

    def NODE(t=0, x=0, y=0, z=0, _from=[], _to=[]):
        if t == 6:
            g = LogicGate(row_start+(-x, y, z), "000000", 2, xaxis=-1, zaxis=2)
            connect(g, g)
        else:
            g = LogicGate(row_start+(-x, y, z), "000000", t, xaxis=-1, zaxis=2)
        connect(_from, g)
        connect(g, _to)
        bp.add(g)
        return g

    g0 = NODE(x=0, _from=execute)
    g1 = NODE(x=1, _from=g0, )
    g2 = NODE(x=2, _from=g1, )
    g3 = NODE(x=3, _from=g2, )
    g4 = NODE(x=4, _from=g3, )
    g5 = NODE(x=5, _from=g4, )
    # If (page loaded)
    g6 = NODE(x=6, _from=(g5), )
    g7 = NODE(x=6, y=1, _from=(cache_table_lookup_out[-1]), )
    g8 = NODE(x=6, y=2, t=3, _from=(cache_table_lookup_out[-1]), )
    # Then:
    g9 = NODE(x=7, _from=(g6, g7), )
    #   If (read mode)
    g10 = NODE(x=8, t=1, _from=(g9), _to=(
        cache_table[1][-1], cache_table[1][-2], cache_table[5]))
    g11 = NODE(x=8, y=1, _from=(read_write_register[0][:, 1]), )
    g12 = NODE(x=8, y=2, t=3, _from=(read_write_register[0][:, 1]), )
    #   Then:
    g13 = NODE(x=9, _from=(g10, g11), _to=(cache[8]))
    g14 = NODE(x=10, _from=(g13))
    g15 = NODE(x=11, _from=(g14))
    g16 = NODE(x=12, _from=(g15))
    g17 = NODE(x=13, _from=(g16), _to=(data_register[1], execution_done))
    #   Else:
    g18 = NODE(x=9, y=2, _from=(g10, g12), _to=(cache[5], execution_done))
    # Else:
    g19 = NODE(x=6, y=3, _from=(g6, g8))
    # g20 = NODE(x=6, y=4, t=1, _from=(g19), _to=(cache_table[8]))
    # g21 = NODE(x=7, y=4, _from=(g20))
    # g22 = NODE(x=8, y=4, _from=(g21))
    # g23 = NODE(x=9, y=4, _from=(g22))
    # g24 = NODE(x=10, y=4, _from=(g23))
    # #   If (Invalid page)
    # g25 = NODE(x=11, y=4, _from=(g24))
    # g26 = NODE(x=11, y=5, _from=(cache_table[2][-2]))
    # g27 = NODE(x=11, y=6, t=3, _from=(cache_table[2][-2]))
    # #   Then:
    # g28 = NODE(x=12, y=4, _from=(g25, g27), _to=(selectors[0]))
    # g29 = NODE(x=13, y=4, _from=(g28), _to=(
    #     cache_table[1][-1], cache_table[1][-2], cache_table[5]))
    # g30 = NODE(x=14, y=4, _from=(g29))
    # g30_1 = NODE(x=15, y=4, _from=(g30), _to=(g10))
    # #   Else:
    # #       If (Visited bit set)
    # g31 = NODE(x=12, y=6, _from=(g25, g26))
    # g31_1 = NODE(x=11, y=8, _from=(cache_table[2][-1]))
    # g32 = NODE(x=12, y=7, _from=(g31_1))
    # g33 = NODE(x=12, y=8, t=3, _from=(g31_1))
    # #       Then:
    # g34 = NODE(x=13, y=6, _from=(g31, g32), _to=(selectors[2], cache_pointer[1]))
    # g35 = NODE(x=14, y=6, _from=(g34), _to=(cache_table[1][-2], cache_table[5]))


    bp.add(cache_arr[:, :, :3], cache_full_page_in, cache_full_page_out,
           timer_bank, timer_bank_decoder_read, timer_bank_decoder_read_in, timer_bank_decoder_read_enable,
           timer_bank_decoder_write_keep, timer_bank_decoder_write_keep_in, timer_bank_decoder_write_keep_enable,
           cache_table_lookup, cache_table_lookup_in, cache_table_lookup_out, cache_table_lookup_dec,
           execute, execution_done, address_comparator, selectors)

    print(
        f"Total memory: {total_bits/8} bytes, "
        f"Cached memory: {cached_bits/8} bytes, ")

    # timers = ndarray((bit_length, 3), dtype=object)
    # timer_read_write_enable = LogicGate(pos + (-1, 1, 0), "FF0000", 4)
    # for x in range(bit_length):
    #     # timers[x] = [
    #     #     t0 := Timer(pos + (x % 8, 0, 2*(x//8)), "000000", ((num_address-3) // 40, (num_address-3) % 40)),
    #     #     l0 := LogicGate(pos + (x % 8, 1, 2*(x//8)), "000000"),
    #     #     l1 := LogicGate(pos + (x % 8, 1, 2*(x//8)+1), "000000", 1),
    #     # ]
    #     timers[x] = [
    #         t0 := Timer(pos + (x, 0, 0), "000000", ((num_address_per_cache-3) // 40, (num_address_per_cache-3) % 40)),
    #         l0 := LogicGate(pos + (x, 1, 0), "000000"),
    #         l1 := LogicGate(pos + (x, 1, 1), "000000", 1),
    #     ]
    #     t0.connect(l0).connect(l1).connect(t0)
    # connect(timer_read_write_enable, timers[:, 1])

    # for z in range(num_caches):
    #     cpos = pos + (0, 0, z*2)
    #     data = register(bp, bit_length, pos=cpos + (0, 4, 0))
    #     timers_readers = [LogicGate(cpos+(x, 3, 0), "000000")
    #                       for x in range(bit_length)]
    #     timers_readers_enable = LogicGate(cpos+(-1, 3, 0), "FF0000", 1)
    #     timers_readers_buffer = LogicGate(cpos+(-1, 5, 0), "000000")
    #     address = register(bp, get_bits_required(
    #         num_address_per_cache), pos=cpos + (bit_length+1, 5, 0), OE=False)
    #     post_rw = register(bp, 2,
    #                        pos=cpos + (bit_length+1+get_bits_required(num_address_per_cache)+1, 5, 0), OE=False)
    #     rw_inv = [
    #         LogicGate(
    #             cpos + (bit_length+1+get_bits_required(num_address_per_cache)+1, 4, 0), "000000", 3),
    #         LogicGate(
    #             cpos + (bit_length+1+get_bits_required(num_address_per_cache)+2, 4, 0), "000000")
    #     ]
    #     comparator = [LogicGate(cpos+(x+bit_length+1, 4, 0), "000000", 5)
    #                   for x in range(get_bits_required(num_address_per_cache))]
    #     comparator_write = LogicGate(cpos+(bit_length+1, 3, 0), "000000")
    #     comparator_read = LogicGate(cpos+(bit_length+2, 3, 0), "000000")
    #     connect(post_rw[0][0, 0], rw_inv)
    #     connect(address[0][:, 0], comparator)
    #     connect(clock[:, 0], comparator)
    #     connect(comparator, comparator_write)
    #     connect(comparator, comparator_read)
    #     connect(rw_inv[0], comparator_write)
    #     connect(rw_inv[1], comparator_read)
    #     connect(post_rw[0][1, 0], comparator_write)
    #     connect(post_rw[0][1, 0], comparator_read)
    #     connect(comparator_write, timer_read_write_enable)
    #     connect(comparator_write, data[2])
    #     connect(data[0][:, 0], timers[:, 2])
    #     connect(comparator_write, post_rw[1])
    #     connect(comparator_read, post_rw[1])
    #     connect(comparator_read, timers_readers_enable)
    #     connect(timers_readers_enable, timers_readers_buffer)
    #     connect(timers_readers_buffer, data[1])
    #     connect(timers_readers_enable, timers_readers)
    #     connect(timers[:, 0], timers_readers)
    #     connect(timers_readers, data[0][:, 3])
    #     bp.add(comparator, comparator_write, comparator_read, rw_inv,
    #            timers_readers, timers_readers_enable, timers_readers_buffer)

    # bp.add(timers, timer_read_write_enable)

    # print(f"Timer ram: {bit_length*num_address_per_cache} bits ({bit_length*num_address_per_cache/8} bytes), "
    #       f"Max time delay: {
    #           num_address_per_cache} ticks ({num_address_per_cache/40} seconds), "
    #       f"Bits per part: {bit_length*num_address_per_cache / (len(bp.bodies[0].childs)-initial_part_count):.2f} bits/part")



from pprint import pp

from numpy import ndarray

from sm_blueprint_lib import Timer, LogicGate, BarrierBlock, Blueprint, save_blueprint, dump_string, connect

size = 10
l0 = [
    LogicGate((x+1, 0, 0), "FF0000", x % 6)
    for x in range(size)
]

l1 = [
    LogicGate((0, x+1, 0), "00FF00", x % 6)
    for x in range(size)
]

l2 = [
    LogicGate((0, 0, x+1), "0000FF", x % 6)
    for x in range(size)
]

# l3 = [
#     [Timer((x+1, y+1, x+y), "000000", (2, 1)) for y in range(size)]
#     for x in range(size)
# ]
l3 = ndarray((size, size), dtype=Timer)
for x in range(size):
    for y in range(size):
        l3[x, y] = Timer((x+1, y+1, x+y), "000000", (2, 1))
base = BarrierBlock((0, -1, -1), "000000", (size+1, size+1, 1))
zero = BarrierBlock((0, -1, 0), "000000", (1, 1, 1))

bp = Blueprint()

connect(l0, l1)
connect(l0, l3)
connect(l3, l2[-1])
connect(l1, l2, parallel=False)
connect(l3.T, l1)

bp.add(l0, l1, l2, base, zero, l3)

# bp.add(l0)
# bp.add(l1)
# bp.add(l2)

# bp.add(base)
# bp.add(zero)
# bp.add(l3)


# bp.bodies[0].childs.extend(l0)
# bp.bodies[0].childs.extend(l1)
# bp.bodies[0].childs.extend(l2)

# bp.bodies[0].childs.append(base)
# bp.bodies[0].childs.append(zero)

# for l in l3:
#     bp.bodies[0].childs.extend(l)

print(len(bp.bodies[0].childs))
path = r"C:\Users\mauri\AppData\Roaming\Axolot Games\Scrap Mechanic\User\User_76561198400983548\Blueprints\c35f6e4e-52cb-4b00-8afa-f0ffd3fbb012\blueprint.json"
save_blueprint(path, bp)



from itertools import batched, cycle
from math import ceil
from typing import Sequence
from numpy import ndarray
from sm_blueprint_lib import get_bits_required, check_pos, connect, num_to_bit_list
from sm_blueprint_lib.blueprint import Blueprint
from sm_blueprint_lib.parts.logicgate import LogicGate
from sm_blueprint_lib.prebuilds.decoder import decoder
from sm_blueprint_lib.parts.timer import Timer
from sm_blueprint_lib.pos import Pos


def rom(
        bp: Blueprint,
        page_size: tuple[int, int],
        data: Sequence[int],
        pos: Pos | Sequence = (0, 0, 0)):
    pos = check_pos(pos)
    data = list(data)
    arr = ndarray((*page_size, 2), dtype=LogicGate)
    page_read = ndarray(page_size[1], dtype=LogicGate)
    page_read2 = ndarray(page_size[1], dtype=LogicGate)
    page_read_binary = ndarray(
        (get_bits_required(page_size[1]), 2), dtype=LogicGate)
    data_out = ndarray(page_size[0], dtype=LogicGate)
    page_writers_binary = ndarray(
        (get_bits_required(len(data)/page_size[1]), 2), dtype=LogicGate)
    page_writers = []
    enable = LogicGate(pos + (-1, page_size[1], 0), "FF00FF", 1)

    for x in range(page_size[0]):
        for y in range(page_size[1]):
            arr[x, y, :] = [
                LogicGate(pos + (x, y, 0), "000000", 0),
                LogicGate(pos + (x, y, 1), "0000FF", 1),
            ]
    page_read[:] = [LogicGate(pos + (-1, y, 0), "000000", 0)
                    for y in range(page_size[1])]
    page_read2[:] = [LogicGate(pos + (-1, y, 1), "000000", 0)
                    for y in range(page_size[1])]
    offset0 = (-get_bits_required(page_size[1]) -
               get_bits_required(len(data)/page_size[1])-1)
    for x in range(get_bits_required(page_size[1])):
        page_read_binary[x] = (
            LogicGate(pos + (x+offset0, page_size[1]-1, 0), "FF0000", 4),
            LogicGate(pos + (x+offset0, page_size[1], 0), "FF0000", 1)
        )
    # page_read_binary[:] = [(LogicGate(pos + (x+offset0, page_size[1]-1, 0), "FF0000", 4),
    #                         LogicGate(pos + (x+offset0, page_size[1], 0), "FF0000", 1))
    #                        for x in range(get_bits_required(page_size[1]))]
    data_out[:] = [LogicGate(pos + (x, page_size[1], 0), "0000FF", 1)
                   for x in range(page_size[0])]

    page_writers_binary[:] = [(LogicGate(pos + (x+offset0+get_bits_required(page_size[1]), page_size[1]-1, 0), "FF0000", 4),
                               LogicGate(pos + (x+offset0+get_bits_required(page_size[1]), page_size[1], 0), "FF0000", 1))
                              for x in range(get_bits_required(len(data)/page_size[1]))]

    for i, data_batch in enumerate(batched(data, page_size[1])):
        g0 = LogicGate(
            pos + (-2-i//(page_size[1]), i % (page_size[1]) - 1, 0), "000000", 0)
        page_writers.append(g0)
        for j, d in enumerate(reversed(data_batch)):
            connect(g0, arr[:, j, 1][num_to_bit_list(d, page_size[0])])

    connect(arr[:, :, 1], arr[:, :, 0])
    connect(arr[:, :, 0], data_out)
    connect(page_read, page_read2)
    connect(page_read2, arr[:, :, 0].T)
    decoder(bp, page_size[1], precreated_inputs_binary=page_read_binary,
            precreated_outputs=list(reversed(page_read)), precreated_output_enable=enable)
    decoder(bp, ceil(len(data)/page_size[1]), precreated_inputs_binary=page_writers_binary,
            precreated_outputs=page_writers, precreated_output_enable=enable, with_enable=False)

    bp.add(arr, page_read, page_read2, page_read_binary, data_out,
           page_writers_binary, page_writers, enable)
    return arr, page_read, page_read_binary, data_out, page_writers_binary, page_writers, enable
